﻿using UnityEngine;
using System.Collections.Generic;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;

/* /////////////////////////////////////////////////////////////////////////////////////////////////

Author: PluginIdea 
Email:819661664@qq.com
Patreon:https://www.patreon.com/PluginIdea

©2022 PluginIdea, Inc. All rights reserved.

///////////////////////////////////////////////////////////////////////////////////////////////// */

namespace PluginIdea
{

    public class MeshDecal : MVRScript
    {

        public struct BlendMode
        {
            public const int Zero = 0;
            public const int One = 1;
            public const int DstColor = 2;
            public const int SrcColor = 3;
            public const int OneMinusDstColor = 4;
            public const int SrcAlpha = 5;
            public const int OneMinusSrcColor = 6;
            public const int DstAlpha = 7;
            public const int OneMinusDstAlpha = 8;
            public const int SrcAlphaSaturate = 9;
            public const int OneMinusSrcAlpha = 10;
        }

        private List<string> blendModes = new List<string> { "Opaque", "AlphaBlend" };
        private List<string> matTypes = new List<string> { "Unlit" };

        private JSONStorableColor mainColor;
        private JSONStorableFloat mainAlpha;
        private JSONStorableFloat glossiness;
        private JSONStorableUrl abedoTextureUrl;
        private JSONStorableUrl normalTextureUrl;
        private JSONStorableFloat metallic;
        private JSONStorableUrl metallicMapUrl;
        private JSONStorableColor emissionColor;
        private JSONStorableFloat cutoutRadio;
        private JSONStorableFloat occlusion;

        private JSONStorableFloat scaleX;
        private JSONStorableFloat scaleY;
        private JSONStorableFloat scaleZ;

        private JSONStorableUrl metallicGlossMap;
        private JSONStorableUrl emissionMapUrl;
        private JSONStorableUrl occlusionMapUrl;

        private JSONStorableBool isShowFrame;

        private JSONStorableStringChooser blendModeChooser;

        private GameObject decalCube;
        private DecalComp decalComp;
        private TimerManager timerManager;

        private Dictionary<string, Material> dic_BundleMaterials = new Dictionary<string, Material>();
        private Dictionary<string, Shader> dic_BundleShaders = new Dictionary<string, Shader>();
        private List<Texture2D> decalTextures;
        private CameraHook cameraHook;

        private float abdeoHWRaido = 1;

        private Material defaultMat;

        private Material mat;
        public Material Mat
        {
            get
            {
                if (mat == null)
                {
                    MeshRenderer meshRenderer = decalComp.MeshHolderTrans.GetComponent<MeshRenderer>();
                    meshRenderer.material = dic_BundleMaterials["DecalMat"];
                    defaultMat = new Material(meshRenderer.material);
                    mat = meshRenderer.material;
                }

                mat.shader = dic_BundleShaders["PluginIdea/UnlitDecal"];
                return mat;
            }
        }

        private bool isExternalResourcesLoaded;
        private bool IsExternalResourcesLoaded
        {
            get { return this.isExternalResourcesLoaded; }
        }

        public bool IsShowWire
        {
            set
            {
                decalComp.WireMeshHolderTrans.gameObject.SetActive(value);
            }
        }

        public override void Init()
        {
            base.Init();

            if (containingAtom.type != "Empty")
            {
                SuperController.LogError("need add on Empty Atom");
                return;
            }

            timerManager = new TimerManager();
            cameraHook = Utils.GetSafeComponent<CameraHook>(Utils.CenterCameraTrans.gameObject);
            cameraHook.onPostRenderCallback += OnPostRender;

            ResourceInit();
            EventInit();
            UIInit();

            decalTextures = new List<Texture2D>() {
                abedoTex, matallicTex, normalTex, occlusionTex, emssionTex
            };
        }

        #region Init
        private void UIInit()
        {
            Utils.OnInitUI(CreateUIElement);
            Utils.SetupInfoText(this, "<size=40><b>Mesh Decal(Free) 1.0</b></size>\n\n" +
                 "<b>ReGenerate Decal:</b>Every time you move the position, you have to regenerate it.\n\n" +
                 "<b>Note: </b>Mesh decal are not real-time. When the contact surface with other mesh objects changes, " +
                 "it needs to be regenerated manually. The advantage of Dynamic decal is :\n\n " +
                 "1: it supports PBR(For Paid Version), once generated,\n" +
                 "2: With higher precision, it will not be effect by shared UV." +
                 "it is equivalent to being completely attached to the surface of the object. Will not affect other objects.", 800, true);

            Utils.SetupSpacer(this, 40, false);
            Utils.SetupButton(this, "ReGenerate Decal", () =>
            {
                SyncDecalSize();
                decalComp.GetTargetObejcts();
            }, false);

            //-------  Color -------------
            //MainColor
            Utils.SetupSpacer(this, 20, true);
            mainColor = Utils.SetupColor(this, "MainColor", Color.white, true);
            mainColor.setCallbackFunction = (h, s, v) =>
            {
                if (!IsExternalResourcesLoaded) return;
                Mat.SetColor("_Color", Color.HSVToRGB(h, s, v));
            };

            emissionColor = Utils.SetupColor(this, "EmissionColor", Color.black, true);
            emissionColor.setCallbackFunction = (h, s, v) =>
            {
                if (!IsExternalResourcesLoaded) return;
                Mat.SetColor("_EmissionColor", Color.HSVToRGB(h, s, v));
            };


            // scale
            Utils.SetupSpacer(this, 60, false);
            scaleX = Utils.SetupSliderFloatWithRange(this, "Scale Width", 0.5f, 0.1f, 5f, false);
            scaleX.setCallbackFunction = (val) =>
            {
                decalCube.transform.localScale = new Vector3(val, decalCube.transform.localScale.y, val * abdeoHWRaido);
            };
            scaleY = Utils.SetupSliderFloatWithRange(this, "Scale Height", 0.5f, 0.1f, 5f, false);
            scaleY.setCallbackFunction = (val) =>
            {
                decalCube.transform.localScale = new Vector3(decalCube.transform.localScale.x, val, decalCube.transform.localScale.z);
            };

            //  "NormalBlend", "Multiply", "SoftAdditive", "LinearDodge"
            blendModeChooser = Utils.SetupStringChooser(this, "BlendMode", blendModes, false);
            blendModeChooser.setCallbackFunction = (val) => {
                if (!IsExternalResourcesLoaded) return;

                cutoutRadio.val = 0.5f;
                switch (val)
                {
                    case "Opaque":
                        Mat.SetInt("_SrcBlend", BlendMode.One);
                        Mat.SetInt("_DstBlend", BlendMode.Zero);
                        break;
                    case "AlphaBlend":
                        Mat.SetInt("_SrcBlend", BlendMode.SrcAlpha);
                        Mat.SetInt("_DstBlend", BlendMode.OneMinusSrcAlpha);
                        cutoutRadio.val = 0;
                        break;
                    default:
                        break;
                }
            };

            ParaMeterSet();
        }

        private Texture2D abedoTex, matallicTex, normalTex, occlusionTex, emssionTex;
        private void ParaMeterSet()
        {
            // common UI
            isShowFrame = Utils.SetupToggle(this, "Show Frame", false, false);

            // wire Mesh
            Utils.SetupToggle(this, "Show Wire Mesh", false, false).setCallbackFunction = (val) =>
            {
                IsShowWire = val;
            };

            // Abedo
            Utils.SetupSpacer(this, 40, false);
            Utils.SetupInfoTextNoScroll(this, "<size=40><b>Common Parameters</b></size>",45, false);
            Utils.SetupInfoTextNoScroll(this, "<size=30><b>Abedo</b></size>", 35, false);
            abedoTextureUrl = Utils.SetupTexture2DChooser(this, "Set Main Decal Texure", $"Custom/Assets/MyAssets", false, new TextureSettings(), (tex) =>
            {
                if (!IsExternalResourcesLoaded) return;

                if (abedoTex != null) Destroy(abedoTex);
                abedoTex = tex;
                Mat.SetTexture("_MainTex", tex);
                SyncDecalSize();
            });

            mainAlpha = Utils.SetupSliderFloatWithRange(this, "Main Alpha", 1, 0f, 1f, false);
            mainAlpha.setCallbackFunction = (val) =>
            {
                if (!IsExternalResourcesLoaded) return;
                Color color = Mat.GetColor("_Color");
                color.a = val;
                Mat.SetColor("_Color", color);
            };

            // cutOut
            cutoutRadio = Utils.SetupSliderFloat(this, "Alpha CutOff", 0.5f, 0f, 1f, false);
            cutoutRadio.setCallbackFunction = (val) =>
            {
                if (!IsExternalResourcesLoaded) return;
                Mat.SetFloat("_CutOff", val);
            };

            Utils.SetupTwinButton(this, "Reload", () => {
                Destroy(Mat.GetTexture("_MainTex"));
                abedoTextureUrl.Reload();
            }, "Remove", () => {
                Destroy(Mat.GetTexture("_MainTex"));
                Mat.SetTexture("_MainTex", defaultMat.GetTexture("_MainTex"));
            }, false);

            // emission
            Utils.SetupSpacer(this, 20, false);
            Utils.SetupInfoTextNoScroll(this, "<size=30><b>Emission</b></size>", 35, false);
            emissionMapUrl = Utils.SetupTexture2DChooser(this, "Set Main EmissionTex", $"Custom/Assets/MyAssets", false, new TextureSettings(), (tex) =>
            {
                if (!IsExternalResourcesLoaded) return;

                if (emssionTex != null) Destroy(emssionTex);
                emssionTex = tex;
                Mat.SetTexture("_EmissionTex", tex);
            });

            Utils.SetupTwinButton(this, "Reload", () => {
                Destroy(Mat.GetTexture("_EmissionTex"));
                emissionMapUrl.Reload();
            }, "Remove", () => {
                Destroy(Mat.GetTexture("_EmissionTex"));
                Mat.SetTexture("_EmissionTex", defaultMat.GetTexture("_EmissionTex"));
            }, false);
        }

        private void ResourceInit()
        {
            OnCreateDecalObject();

            string assetsBundlePath = Utils.GetPluginPath(this) + "/resources/meshdecal.assetbundle";
            Request request = new Request { path = assetsBundlePath, callback = OnAssetBundleLoaded };
            AssetLoader.QueueLoadAssetBundleFromFile(request);
        }

        private void OnAssetBundleLoaded(Request request)
        {
            Material[] materials = request.assetBundle.LoadAllAssets<Material>();
            for (int i = 0; i < materials.Length; i++)
            {
                //SuperController.LogMessage("Mat: " + materials[i].name);
                if (!dic_BundleMaterials.ContainsKey(materials[i].name))
                {
                    dic_BundleMaterials.Add(materials[i].name, materials[i]);
                }
            }

            Shader[] shaders = request.assetBundle.LoadAllAssets<Shader>();
            for (int i = 0; i < shaders.Length; i++)
            {
                //SuperController.LogMessage("Shader: " + shaders[i].name);
                if (!dic_BundleShaders.ContainsKey(shaders[i].name))
                {
                    dic_BundleShaders.Add(shaders[i].name, shaders[i]);
                }
            }

            // reload
            // abedoTex, matallicTex, normalTex, occlusionTex, emssionTex;
            isExternalResourcesLoaded = true;
            abedoTextureUrl.Reload();
            emissionMapUrl.Reload();

            // value trigger
            mainAlpha.setCallbackFunction(mainAlpha.val);
            cutoutRadio.setCallbackFunction(cutoutRadio.val);
            mainColor.setCallbackFunction(mainColor.val.H, mainColor.val.S, mainColor.val.V);
            emissionColor.setCallbackFunction(emissionColor.val.H, emissionColor.val.S, emissionColor.val.V);
            blendModeChooser.setCallbackFunction(blendModeChooser.val);

            // wire mesh
            decalComp.WireMeshHolderTrans.GetComponent<MeshRenderer>().material = dic_BundleMaterials["WireLine"];
        }

        private void EventInit()
        {
            SuperController.singleton.onSceneLoadedHandlers += OnAfterSceneLoaded;
        }
        #endregion

        #region callback
        private void OnAfterSceneLoaded()
        {
            timerManager.AddTimer(0.5f, () => {
                decalComp.GetTargetObejcts();
            });
        }
        #endregion

        private void Update()
        {
            timerManager.Update();
        }

        private void OnCreateDecalObject()
        {
            decalCube = null;
            decalCube = GameObject.CreatePrimitive(PrimitiveType.Cube);
            decalCube.GetComponent<Collider>().enabled = false;
            decalCube.GetComponent<MeshRenderer>().enabled = false;
            decalCube.transform.SetParent(containingAtom.reParentObject.Find("object/rescaleObject"));
            decalCube.transform.localScale = new Vector3(0.5f, 0.5f, 0.5f);
            decalCube.transform.localPosition = Vector3.zero;
            decalCube.transform.localRotation = Quaternion.identity;
            decalComp = Utils.GetSafeComponent<DecalComp>(decalCube);
            decalComp.Init(this);
        }

        private void SyncDecalSize()
        {
            if (abedoTex == null) return;

            abdeoHWRaido = (float)abedoTex.height / abedoTex.width;
            decalCube.transform.localScale = new Vector3(decalCube.transform.localScale.x, decalCube.transform.localScale.x * abdeoHWRaido, decalCube.transform.localScale.z);
        }

        private Color gizmosUpColor = Color.gray;
        private Color gizmosDownColor = Color.green;
        private void OnPostRender()
        {
            if (decalCube != null && isShowFrame.val)
            {
                Transform decalTrans = decalCube.transform;

                Vector3 u1 = decalTrans.TransformPoint(new Vector3(-0.5f, 0.5f, 0.5f));
                Vector3 u2 = decalTrans.TransformPoint(new Vector3(0.5f, 0.5f, 0.5f));
                Vector3 u3 = decalTrans.TransformPoint(new Vector3(0.5f, -0.5f, 0.5f));
                Vector3 u4 = decalTrans.TransformPoint(new Vector3(-0.5f, -0.5f, 0.5f));

                Vector3 d1 = decalTrans.TransformPoint(new Vector3(-0.5f, 0.5f, -0.5f));
                Vector3 d2 = decalTrans.TransformPoint(new Vector3(0.5f, 0.5f, -0.5f));
                Vector3 d3 = decalTrans.TransformPoint(new Vector3(0.5f, -0.5f, -0.5f));
                Vector3 d4 = decalTrans.TransformPoint(new Vector3(-0.5f, -0.5f, -0.5f));

                GLDrawHelper.DrawLine3D(new List<GLLine3D> {
                    new GLLine3D(){ from = u1, to = u2, color = gizmosUpColor},
                    new GLLine3D(){ from = u2, to = u3, color = gizmosUpColor},
                    new GLLine3D(){ from = u3, to = u4, color = gizmosUpColor},
                    new GLLine3D(){ from = u4, to = u1, color = gizmosUpColor},
                }, null);

                GLDrawHelper.DrawLine3D(new List<GLLine3D> {
                    new GLLine3D(){ from = d1, to = d2, color = gizmosDownColor},
                    new GLLine3D(){ from = d2, to = d3, color = gizmosDownColor},
                    new GLLine3D(){ from = d3, to = d4, color = gizmosDownColor},
                    new GLLine3D(){ from = d4, to = d1, color = gizmosDownColor},
                }, null);

                GLDrawHelper.DrawLine3D(new List<GLLine3D> {
                    new GLLine3D(){ from = u1, to = d1, color = Color.white},
                    new GLLine3D(){ from = u2, to = d2, color = Color.white},
                    new GLLine3D(){ from = u3, to = d3, color = Color.white},
                    new GLLine3D(){ from = u4, to = d4, color = Color.white},
                }, null);

                GizmosHelper.DrawArrow3D(decalTrans, decalCube.transform.localScale.z * 0.4f, Color.green);
            }
        }

        private void OnDestroy()
        {
            if (decalCube != null)
            {
                Destroy(decalCube);
            }
            timerManager = null;
            Destroy(cameraHook);
            SuperController.singleton.onSceneLoadedHandlers -= OnAfterSceneLoaded;

            // destroy textures
            decalTextures.ForEach(tex => {
                if (tex != null)
                {
                    Destroy(tex);
                }
            });
            decalTextures.Clear();

            Utils.OnDestroyUI();
        }
    }
}